var CssRender = (function (exports) {
    'use strict';

    function ampCount(selector) {
        let cnt = 0;
        for (let i = 0; i < selector.length; ++i) {
            if (selector[i] === '&')
                ++cnt;
        }
        return cnt;
    }
    /**
     * Don't just use ',' to separate css selector. For example:
     * x:(a, b) {} will be split into 'x:(a' and 'b)', which is not expected.
     * Make sure comma doesn't exist inside parentheses.
     */
    const seperatorRegex = /\s*,(?![^(]*\))\s*/g;
    const extraSpaceRegex = /\s+/g;
    /**
     * selector must includes '&'
     * selector is trimmed
     * every part of amp is trimmed
     */
    function resolveSelectorWithAmp(amp, selector) {
        const nextAmp = [];
        selector.split(seperatorRegex).forEach(partialSelector => {
            let round = ampCount(partialSelector);
            // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
            if (!round) {
                amp.forEach(partialAmp => {
                    nextAmp.push(
                    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
                    (partialAmp && partialAmp + ' ') + partialSelector);
                });
                return;
            }
            else if (round === 1) {
                amp.forEach(partialAmp => {
                    nextAmp.push(partialSelector.replace('&', partialAmp));
                });
                return;
            }
            let partialNextAmp = [
                partialSelector
            ];
            // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
            while (round--) {
                const nextPartialNextAmp = [];
                partialNextAmp.forEach(selectorItr => {
                    amp.forEach(partialAmp => {
                        nextPartialNextAmp.push(selectorItr.replace('&', partialAmp));
                    });
                });
                partialNextAmp = nextPartialNextAmp;
            }
            partialNextAmp.forEach(part => nextAmp.push(part));
        });
        return nextAmp;
    }
    /**
     * selector mustn't includes '&'
     * selector is trimmed
     */
    function resolveSelector(amp, selector) {
        const nextAmp = [];
        selector.split(seperatorRegex).forEach(partialSelector => {
            amp.forEach(partialAmp => {
                // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
                nextAmp.push(((partialAmp && partialAmp + ' ') + partialSelector));
            });
        });
        return nextAmp;
    }
    function parseSelectorPath(selectorPaths) {
        let amp = [''];
        selectorPaths.forEach(selector => {
            // eslint-disable-next-line
            selector = selector && selector.trim();
            if (
            // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
            !selector) {
                /**
                 * if it's a empty selector, do nothing
                 */
                return;
            }
            // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
            if (selector.includes('&')) {
                amp = resolveSelectorWithAmp(amp, selector);
            }
            else {
                amp = resolveSelector(amp, selector);
            }
        });
        return amp.join(', ').replace(extraSpaceRegex, ' ');
    }

    const kebabRegex = /[A-Z]/g;
    function kebabCase(pattern) {
        return pattern.replace(kebabRegex, match => '-' + match.toLowerCase());
    }
    /** TODO: refine it to solve nested object */
    function upwrapProperty(prop, indent = '  ') {
        if (typeof prop === 'object' && prop !== null) {
            return (' {\n' +
                Object.entries(prop).map(v => {
                    return indent + `  ${kebabCase(v[0])}: ${v[1]};`;
                }).join('\n') +
                '\n' + indent + '}');
        }
        return `: ${prop};`;
    }
    /** unwrap properties */
    function upwrapProperties(props, instance, params) {
        if (typeof props === 'function') {
            return props({
                context: instance.context,
                props: params
            });
        }
        return props;
    }
    function createStyle(selector, props, instance, params) {
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (!props)
            return '';
        // eslint-disable-next-line
        const unwrappedProps = upwrapProperties(props, instance, params);
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (!unwrappedProps)
            return '';
        const propertyNames = Object.keys(unwrappedProps);
        if (propertyNames.length === 0) {
            // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
            if (instance.config.keepEmptyBlock)
                return selector + ' {\n}';
            return '';
        }
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        const statements = selector ? [
            selector + ' {'
        ] : [];
        propertyNames.forEach(propertyName => {
            const property = unwrappedProps[propertyName];
            if (propertyName === 'raw') {
                statements.push('\n' + property + '\n');
                return;
            }
            propertyName = kebabCase(propertyName);
            // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
            if (property !== null && property !== undefined) {
                statements.push(`  ${propertyName}${upwrapProperty(property)}`);
            }
        });
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (selector) {
            statements.push('}');
        }
        return statements.join('\n');
    }
    function loopCNodeListWithCallback(children, options, callback) {
        /* istanbul ignore if */
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (!children)
            return;
        children.forEach(child => {
            if (Array.isArray(child)) {
                loopCNodeListWithCallback(child, options, callback);
            }
            else if (typeof child === 'function') {
                const grandChildren = child(options);
                if (Array.isArray(grandChildren)) {
                    loopCNodeListWithCallback(grandChildren, options, callback);
                    // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
                }
                else if (grandChildren) {
                    callback(grandChildren);
                }
                // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
            }
            else if (child) {
                callback(child);
            }
        });
    }
    function traverseCNode(node, selectorPaths, styles, instance, params) {
        const $ = node.$;
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (!$ || typeof $ === 'string') {
            // as a string selector
            selectorPaths.push($);
        }
        else if (typeof $ === 'function') {
            // as a lazy selector
            selectorPaths.push($({
                context: instance.context,
                props: params
            }));
        }
        else { // as a option selector
            // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
            if ($.before)
                $.before(instance.context);
            // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
            if (!$.$ || typeof $.$ === 'string') {
                selectorPaths.push($.$);
                // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
            }
            else /* istanbul ignore else */ if ($.$) {
                selectorPaths.push($.$({
                    context: instance.context,
                    props: params
                }));
            }
        }
        const selector = parseSelectorPath(selectorPaths);
        const style = createStyle(selector, node.props, instance, params);
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (style.length)
            styles.push(style);
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (node.children) {
            loopCNodeListWithCallback(node.children, {
                context: instance.context,
                props: params
            }, childNode => {
                if (typeof childNode === 'string') {
                    styles.push(createStyle(selector, { raw: childNode }, instance, params));
                }
                else {
                    traverseCNode(childNode, selectorPaths, styles, instance, params);
                }
            });
        }
        selectorPaths.pop();
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if ($ && $.after)
            $.after(instance.context);
    }
    function render(node, instance, props) {
        const styles = [];
        traverseCNode(node, [], styles, instance, props);
        return styles.join('\n\n');
    }

    function removeElement(el) {
        /* istanbul ignore if */
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (!el)
            return;
        const parentElement = el.parentElement;
        /* istanbul ignore else */
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (parentElement)
            parentElement.removeChild(el);
    }
    function queryElement(id) {
        return document.querySelector(`style[cssr-id="${String(id)}"]`);
    }
    function createElement(id) {
        const el = document.createElement('style');
        el.setAttribute('cssr-id', String(id));
        return el;
    }

    function getCountOfElement(el) {
        const count = el.getAttribute('mount-count');
        if (count === null)
            return 0;
        return Number(count);
    }
    function setCountOfElement(el, count) {
        if (count === null) {
            el.removeAttribute('mount-count');
        }
        else {
            el.setAttribute('mount-count', String(count));
        }
    }
    function unmount(intance, node, target, count) {
        const els = node.els;
        if (target === undefined) {
            els.forEach(removeElement);
            node.els = [];
        }
        else if (typeof target === 'string' || typeof target === 'number') {
            const targetElement = queryElement(target);
            // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
            if (targetElement && els.includes(targetElement)) {
                const mountCount = getCountOfElement(targetElement);
                if (!count || mountCount <= 1) {
                    removeElement(targetElement);
                    node.els = els.filter(el => el !== targetElement);
                }
                else {
                    setCountOfElement(targetElement, mountCount - 1);
                }
            }
        }
        else {
            const mountCount = getCountOfElement(target);
            if (!count || mountCount <= 1) {
                setCountOfElement(target, null);
                target.innerHTML = '';
            }
            else {
                setCountOfElement(target, mountCount - 1);
            }
        }
    }
    function addElementToList(els, target) {
        if (!els.includes(target)) {
            els.push(target);
        }
    }
    function mount(instance, node, target, props, count) {
        let targetElement = null;
        const els = node.els;
        if (target === undefined) {
            targetElement = document.createElement('style');
            document.head.appendChild(targetElement);
            addElementToList(els, targetElement);
        }
        else if (typeof target === 'string' || typeof target === 'number') {
            targetElement = queryElement(target);
            if (targetElement === null) {
                targetElement = createElement(target);
                document.head.appendChild(targetElement);
                if (count) {
                    setCountOfElement(targetElement, 1);
                }
                addElementToList(els, targetElement);
            }
            else {
                if (count) {
                    setCountOfElement(targetElement, getCountOfElement(targetElement) + 1);
                }
                addElementToList(els, targetElement);
                return targetElement;
            }
        }
        else {
            targetElement = target;
            const mountCount = getCountOfElement(targetElement);
            if (mountCount > 0) {
                if (count) {
                    setCountOfElement(targetElement, mountCount + 1);
                }
                return target;
            }
            else if (count) {
                setCountOfElement(targetElement, 1);
            }
        }
        const style = node.render(props);
        if (targetElement.innerHTML !== style) {
            targetElement.innerHTML = style;
        }
        return targetElement;
    }

    function wrappedRender(props) {
        return render(this, this.instance, props);
    }
    const inNode = typeof window === 'undefined';
    function wrappedMount(options) {
        /* istanbul ignore next */
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (inNode)
            return null;
        // eslint-disable-next-line
        const target = options && options.target;
        if (target === null)
            return null;
        // eslint-disable-next-line
        const props = options && options.props;
        // eslint-disable-next-line
        const count = !((options && options.count) === false);
        const targetElement = mount(this.instance, this, target, props, count);
        return targetElement;
    }
    function wrappedUnmount(options) {
        /* istanbul ignore next */
        // eslint-disable-next-line @typescript-eslint/strict-boolean-expressions
        if (inNode)
            return;
        // eslint-disable-next-line
        const target = options && options.target;
        // eslint-disable-next-line
        const delay = (options && options.delay) || 0;
        // eslint-disable-next-line
        const count = !((options && options.count) === false);
        if (target === null)
            return;
        if (delay === 0)
            unmount(this.instance, this, target, count);
        else {
            setTimeout(() => unmount(this.instance, this, target, count), delay);
        }
    }
    const createCNode = function (instance, $, props, children) {
        return {
            instance,
            $,
            props,
            children,
            els: [],
            render: wrappedRender,
            mount: wrappedMount,
            unmount: wrappedUnmount
        };
    };
    const c = function (instance, $, props, children) {
        if (Array.isArray($)) {
            return createCNode(instance, { $: null }, null, $);
        }
        if (Array.isArray(props)) {
            return createCNode(instance, $, null, props);
        }
        else if (Array.isArray(children)) {
            return createCNode(instance, $, props, children);
        }
        else {
            return createCNode(instance, $, props, null);
        }
    };

    function CssRender(config = {}) {
        const cssr = {
            c: ((...args) => c(cssr, ...args)),
            use: (plugin, ...args) => plugin.install(cssr, ...args),
            find: queryElement,
            context: {},
            config
        };
        return cssr;
    }

    exports.CssRender = CssRender;
    exports.default = CssRender;

    return exports;

}({}));
